import os
import sqlite3

from dataclasses import dataclass

from mangum.backends.base import WebSocketBackend
from mangum.exceptions import ConfigurationError


@dataclass
class SQLiteBackend(WebSocketBackend):
    def __post_init__(self) -> None:
        try:
            file_path = self.params["file_path"]
        except KeyError:
            raise ConfigurationError(f"SQLite3 database 'file_path' missing.")
        self.table_name = self.params.get("table_name", "mangum")
        if not os.path.exists(file_path):
            self.connection = sqlite3.connect(file_path)
            self.cursor = self.connection.cursor()
            self.cursor.execute(
                f"create table {self.table_name} (id varchar(64) primary key, initial_scope text)"
            )
        else:
            self.connection = sqlite3.connect(file_path)
            self.cursor = self.connection.cursor()

    def create(self, connection_id: str, initial_scope: str) -> None:
        self.cursor.execute(
            f"insert into {self.table_name} values (?, ?)",
            (connection_id, initial_scope),
        )
        self.connection.commit()
        self.cursor.close()
        self.connection.close()

    def fetch(self, connection_id: str) -> str:
        initial_scope = self.cursor.execute(
            f"select initial_scope from {self.table_name} where id = ?",
            (connection_id,),
        ).fetchone()[0]
        self.cursor.close()
        self.connection.close()

        return initial_scope

    def delete(self, connection_id: str) -> None:
        self.cursor.execute(
            f"delete from {self.table_name} where id = ?", (connection_id,)
        ).fetchone()
        self.connection.commit()
        self.cursor.close()
        self.connection.close()



import base64
import asyncio
import urllib.parse
import typing
import logging
import os
from dataclasses import dataclass

from mangum.lifespan import Lifespan
from mangum.types import ASGIApp
from mangum.protocols.http import HTTPCycle
from mangum.protocols.ws import WebSocketCycle
from mangum.websocket import WebSocket
from mangum.exceptions import ConfigurationError


DEFAULT_TEXT_MIME_TYPES = [
    "application/json",
    "application/javascript",
    "application/xml",
    "application/vnd.api+json",
]


def get_server(headers: dict) -> typing.Tuple:  # pragma: no cover
    server_name = headers.get("host", "mangum")
    if ":" not in server_name:
        server_port = headers.get("x-forwarded-port", 80)
    else:
        server_name, server_port = server_name.split(":")
    server = (server_name, int(server_port))

    return server


def get_logger(log_level: str) -> logging.Logger:
    level = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG,
    }[log_level]
    logging.basicConfig(
        format="[%(asctime)s] %(message)s", level=level, datefmt="%d-%b-%y %H:%M:%S"
    )
    logger = logging.getLogger("mangum")
    logger.setLevel(level)

    return logger


@dataclass
class Mangum:

    app: ASGIApp
    enable_lifespan: bool = True
    log_level: str = "info"
    api_gateway_base_path: typing.Optional[str] = None
    text_mime_types: typing.Optional[typing.List[str]] = None
    ws_config: typing.Optional[dict] = None

    def __post_init__(self) -> None:
        self.logger = get_logger(self.log_level)
        if self.enable_lifespan:
            loop = asyncio.get_event_loop()
            self.lifespan = Lifespan(self.app)
            loop.create_task(self.lifespan.run())
            loop.run_until_complete(self.lifespan.startup())

    def __call__(self, event: dict, context: dict) -> dict:
        response = self.handler(event, context)

        return response

    def strip_base_path(self, path: str) -> str:
        if self.api_gateway_base_path:
            script_name = "/" + self.api_gateway_base_path
            if path.startswith(script_name):
                path = path[len(script_name) :]

        return urllib.parse.unquote(path or "/")

    def handler(self, event: dict, context: dict) -> dict:
        if "eventType" in event["requestContext"]:
            response = self.handle_ws(event, context)
        else:
            is_http_api = "http" in event["requestContext"]
            response = self.handle_http(event, context, is_http_api=is_http_api)

        if self.enable_lifespan:
            if self.lifespan.is_supported:
                loop = asyncio.get_event_loop()
                loop.run_until_complete(self.lifespan.shutdown())

        return response

    def handle_http(self, event: dict, context: dict, *, is_http_api: bool) -> dict:
        self.logger.info("HTTP event received.")
        if is_http_api:
            source_ip = event["requestContext"]["http"]["sourceIp"]
            query_string = event.get("rawQueryString")
            path = event["requestContext"]["http"]["path"]
            http_method = event["requestContext"]["http"]["method"]
        else:
            source_ip = event["requestContext"].get("identity", {}).get("sourceIp")
            multi_value_query_string_params = event["multiValueQueryStringParameters"]
            query_string = (
                urllib.parse.urlencode(
                    multi_value_query_string_params, doseq=True
                ).encode()
                if multi_value_query_string_params
                else b""
            )
            path = event["path"]
            http_method = event["httpMethod"]

        headers = (
            {k.lower(): v for k, v in event.get("headers").items()}  # type: ignore
            if event.get("headers")
            else {}
        )
        server = get_server(headers)
        client = (source_ip, 0)

        scope = {
            "type": "http",
            "http_version": "1.1",
            "method": http_method,
            "headers": [[k.encode(), v.encode()] for k, v in headers.items()],
            "path": self.strip_base_path(path),
            "raw_path": None,
            "root_path": "",
            "scheme": headers.get("x-forwarded-proto", "https"),
            "query_string": query_string,
            "server": server,
            "client": client,
            "asgi": {"version": "3.0"},
            "aws.event": event,
            "aws.context": context,
        }

        is_binary = event.get("isBase64Encoded", False)
        body = event.get("body") or b""
        if is_binary:
            body = base64.b64decode(body)
        elif not isinstance(body, bytes):
            body = body.encode()

        if self.text_mime_types:
            text_mime_types = self.text_mime_types + DEFAULT_TEXT_MIME_TYPES
        else:
            text_mime_types = DEFAULT_TEXT_MIME_TYPES

        asgi_cycle = HTTPCycle(
            scope, text_mime_types=text_mime_types, log_level=self.log_level
        )
        asgi_cycle.put_message(
            {"type": "http.request", "body": body, "more_body": False}
        )
        response = asgi_cycle(self.app)

        return response

    def handle_ws(self, event: dict, context: dict) -> dict:
        if self.ws_config is None:
            raise ConfigurationError(
                "A `ws_config` configuration mapping is required for WebSocket support."
            )

        event_type = event["requestContext"]["eventType"]
        connection_id = event["requestContext"]["connectionId"]
        stage = event["requestContext"]["stage"]
        domain_name = event["requestContext"]["domainName"]

        api_gateway_endpoint_url = self.ws_config.get(
            "api_gateway_endpoint_url", f"https://{domain_name}/{stage}"
        )
        api_gateway_region_name = self.ws_config.get(
            "api_gateway_region_name", os.environ["AWS_REGION"]
        )

        websocket = WebSocket(
            connection_id,
            ws_config=self.ws_config,
            api_gateway_endpoint_url=api_gateway_endpoint_url,
            api_gateway_region_name=api_gateway_region_name,
        )

        self.logger.info(
            "%s event received for WebSocket connection %s", event_type, connection_id
        )
        if event_type == "CONNECT":

            headers = (
                {k.lower(): v for k, v in event.get("headers").items()}  # type: ignore
                if event.get("headers")
                else {}
            )
            server = get_server(headers)
            source_ip = event["requestContext"].get("identity", {}).get("sourceIp")
            client = (source_ip, 0)

            initial_scope = {
                "type": "websocket",
                "path": "/",
                "headers": headers,
                "raw_path": None,
                "root_path": "",
                "scheme": headers.get("x-forwarded-proto", "wss"),
                "query_string": "",
                "server": server,
                "client": client,
                "aws": {"event": event, "context": None},
            }

            websocket.create(initial_scope)
            response = {"statusCode": 200}

        elif event_type == "MESSAGE":
            websocket.fetch()
            asgi_cycle = WebSocketCycle(websocket, log_level=self.log_level)
            asgi_cycle.put_message({"type": "websocket.connect"})
            asgi_cycle.put_message(
                {"type": "websocket.receive", "bytes": None, "text": event["body"]}
            )

            response = asgi_cycle(self.app)

        elif event_type == "DISCONNECT":
            websocket.delete()
            response = {"statusCode": 200}

        return response



import os
import logging
from dataclasses import dataclass

import boto3
from botocore.client import Config
from botocore.exceptions import ClientError

from mangum.backends.base import WebSocketBackend
from mangum.exceptions import ConfigurationError


@dataclass
class S3Backend(WebSocketBackend):
    def __post_init__(self) -> None:
        self.logger = logging.getLogger("mangum.websocket.s3")
        try:
            self.bucket = self.params["bucket"]
        except KeyError:
            raise ConfigurationError("S3 'bucket' parameter missing.")
        region_name = self.params.get("region_name", os.environ["AWS_REGION"])
        self.connection = boto3.client(
            "s3",
            region_name=region_name,
            config=Config(connect_timeout=5, retries={"max_attempts": 0}),
        )
        create_bucket = False
        try:
            self.connection.head_bucket(Bucket=self.bucket)
        except ClientError as exc:
            error_code = int(exc.response["Error"]["Code"])
            if error_code == 403:  # pragma: no cover
                self.logger.error("S3 bucket access forbidden!")
            elif error_code == 404:
                self.logger.info(f"Bucket {self.bucket} not found, creating.")
                create_bucket = True

        if create_bucket:
            self.connection.create_bucket(
                Bucket=self.bucket,
                CreateBucketConfiguration={"LocationConstraint": region_name},
            )

    def create(self, connection_id: str, initial_scope: str) -> None:
        self.connection.put_object(
            Body=initial_scope.encode(), Bucket=self.bucket, Key=connection_id
        )

    def fetch(self, connection_id: str) -> str:
        s3_object = self.connection.get_object(Bucket=self.bucket, Key=connection_id)
        initial_scope = s3_object["Body"].read().decode()

        return initial_scope

    def delete(self, connection_id: str) -> None:
        self.connection.delete_object(Bucket=self.bucket, Key=connection_id)



class LifespanFailure(Exception):
    """Raise when an error occurs in a lifespan event"""


class WebSocketError(Exception):
    """Raise when an error occurs in a WebSocket event."""


class ConfigurationError(Exception):
    """Raise when an error occurs parsing configuration."""

    # def __init__(self, param: str) -> None:
    #     super().__init__(f"Missing {param} in configuration")



import os
from dataclasses import dataclass

import boto3
from botocore.exceptions import ClientError

from mangum.backends.base import WebSocketBackend
from mangum.exceptions import WebSocketError, ConfigurationError


@dataclass
class DynamoDBBackend(WebSocketBackend):
    def __post_init__(self) -> None:
        try:
            table_name = self.params["table_name"]
        except KeyError:
            raise ConfigurationError("DynamoDB 'table_name' missing.")
        region_name = self.params.get("region_name", os.environ["AWS_REGION"])
        endpoint_url = self.params.get("endpoint_url", None)
        try:
            dynamodb_resource = boto3.resource(
                "dynamodb", region_name=region_name, endpoint_url=endpoint_url
            )
            dynamodb_resource.meta.client.describe_table(TableName=table_name)
        except ClientError as exc:
            raise WebSocketError(exc)

        self.connection = dynamodb_resource.Table(table_name)

    def create(self, connection_id: str, initial_scope: str) -> None:
        self.connection.put_item(
            Item={"connectionId": connection_id, "initial_scope": initial_scope},
            ConditionExpression=f"attribute_not_exists(connectionId)",
        )

    def fetch(self, connection_id: str) -> str:
        try:
            item = self.connection.get_item(Key={"connectionId": connection_id})["Item"]
        except KeyError:
            raise WebSocketError(f"Connection not found: {connection_id}")

        initial_scope = item["initial_scope"]

        return initial_scope

    def delete(self, connection_id: str) -> None:
        self.connection.delete_item(Key={"connectionId": connection_id})



import logging
from dataclasses import dataclass

import psycopg2
from psycopg2 import sql

from mangum.backends.base import WebSocketBackend
from mangum.exceptions import ConfigurationError


@dataclass
class PostgreSQLBackend(WebSocketBackend):
    def __post_init__(self) -> None:
        self.logger = logging.getLogger("mangum.websocket.postgres")
        self.logger.debug("Connecting to PostgreSQL database.")
        connect_timeout = self.params.get("connect_timeout", 5)
        if "uri" in self.params:
            self.connection = psycopg2.connect(
                self.params["uri"], connect_timeout=connect_timeout
            )
        else:
            try:
                database = self.params["database"]
                user = self.params["user"]
                password = self.params["password"]
                host = self.params["host"]
            except KeyError:  # pragma: no cover
                raise ConfigurationError("PostgreSQL connection details missing.")
            port = self.params.get("port", "5432")  # pragma: no cover
            self.connection = psycopg2.connect(
                database=database,
                user=user,
                password=password,
                host=host,
                port=port,
                connect_timeout=connect_timeout,
            )
        self.table_name = self.params.get("table_name", "mangum")
        self.cursor = self.connection.cursor()
        self.cursor.execute(
            sql.SQL(
                "create table if not exists {} (id varchar(64) primary key, initial_scope text)"
            ).format(sql.Identifier(self.table_name))
        )
        self.connection.commit()
        self.logger.debug("Connection established.")

    def create(self, connection_id: str, initial_scope: str) -> None:
        self.logger.debug("Creating database entry for %s", connection_id)
        self.cursor.execute(
            sql.SQL("insert into {} values (%s, %s)").format(
                sql.Identifier(self.table_name)
            ),
            (connection_id, initial_scope),
        )

        self.connection.commit()
        self.connection.close()
        self.logger.debug("Database entry created.")

    def fetch(self, connection_id: str) -> str:
        self.logger.debug("Fetching initial scope for %s", connection_id)
        self.cursor.execute(
            sql.SQL("select initial_scope from {} where id = %s").format(
                sql.Identifier(self.table_name)
            ),
            (connection_id,),
        )
        initial_scope = self.cursor.fetchone()[0]
        self.cursor.close()
        self.connection.close()
        self.logger.debug("Initial scope fetched.")

        return initial_scope

    def delete(self, connection_id: str) -> None:
        self.logger.debug("Deleting database entry for %s", connection_id)
        self.cursor.execute(
            sql.SQL("delete from {} where id = %s").format(
                sql.Identifier(self.table_name)
            ),
            (connection_id,),
        )
        self.connection.commit()
        self.cursor.close()
        self.connection.close()
        self.logger.debug("Database entry deleted.")



from dataclasses import dataclass


@dataclass
class WebSocketBackend:
    """
    Base class for implementing WebSocket backends to store APIGateway connections.
    """

    params: dict

    def create(self, connection_id: str, initial_scope: str) -> None:
        """
        Store the connection id and initial scope during the WebSocket CONNECT event.
        """
        raise NotImplementedError()

    def fetch(self, connection_id: str) -> str:
        """
        Retrieve and return the initial scope during the WebSocket MESSAGE event.
        """
        raise NotImplementedError()

    def delete(self, connection_id: str) -> None:
        """
        Delete the stored connection during the WebSocket DISCONNECT event or when
        a stale connection is detected in API Gateway.
        """
        raise NotImplementedError()



import logging
from dataclasses import dataclass

import redis

from mangum.backends.base import WebSocketBackend
from mangum.exceptions import ConfigurationError


@dataclass
class RedisBackend(WebSocketBackend):
    def __post_init__(self) -> None:
        self.logger = logging.getLogger("mangum.websocket.redis")
        self.logger.debug("Connecting to Redis host.")
        try:
            host = self.params["host"]
        except KeyError:  # pragma: no cover
            raise ConfigurationError("Redis 'host' parameter missing.")
        password = self.params.get("password")
        port = self.params.get("port", "5432")  # pragma: no cover
        self.connection = redis.Redis(host, port, password=password)
        self.logger.debug("Connection established.")

    def create(self, connection_id: str, initial_scope: str) -> None:
        self.logger.debug("Creating entry for %s", connection_id)
        self.connection.set(connection_id, initial_scope)
        self.logger.debug("Entry created.")

    def fetch(self, connection_id: str) -> str:
        self.logger.debug("Fetching initial scope for %s", connection_id)
        initial_scope = self.connection.get(connection_id)
        self.logger.debug("Initial scope fetched.")

        return initial_scope

    def delete(self, connection_id: str) -> None:
        self.logger.debug("Deleting entry for %s", connection_id)
        self.connection.delete(connection_id)
        self.logger.debug("Entry deleted.")



import asyncio
import logging
from dataclasses import dataclass

from mangum.types import ASGIApp, Message
from mangum.exceptions import LifespanFailure


@dataclass
class Lifespan:
    app: ASGIApp
    is_supported: bool = False
    has_error: bool = False

    def __post_init__(self) -> None:
        self.logger: logging.Logger = logging.getLogger("mangum.lifespan")
        self.app_queue: asyncio.Queue = asyncio.Queue()
        self.startup_event: asyncio.Event = asyncio.Event()
        self.shutdown_event: asyncio.Event = asyncio.Event()

    async def startup(self) -> None:
        if self.is_supported:
            self.logger.info("Waiting for application startup.")
            await self.app_queue.put({"type": "lifespan.startup"})
            await self.startup_event.wait()
            if self.has_error:
                self.logger.error("Application startup failed.")
            else:
                self.logger.info("Application startup complete.")

    async def shutdown(self) -> None:
        if self.has_error:
            return
        self.logger.info("Waiting for application shutdown.")
        await self.app_queue.put({"type": "lifespan.shutdown"})
        await self.shutdown_event.wait()

    async def run(self) -> None:
        try:
            await self.app({"type": "lifespan"}, self.receive, self.send)
        except BaseException as exc:
            self.startup_event.set()
            self.shutdown_event.set()
            self.has_error = True
            if not self.is_supported:
                self.logger.info("ASGI 'lifespan' protocol appears unsupported.")
            else:
                self.logger.error("Exception in 'lifespan' protocol.", exc_info=exc)

    async def receive(self) -> Message:
        self.is_supported = True

        return await self.app_queue.get()

    async def send(self, message: Message) -> None:
        if not self.is_supported:
            raise LifespanFailure("Lifespan unsupported.")

        message_type = message["type"]
        if message_type == "lifespan.startup.complete":
            self.startup_event.set()
        elif message_type == "lifespan.shutdown.complete":
            self.shutdown_event.set()
        else:
            raise RuntimeError(
                f"Expected lifespan message type, received: {message_type}"
            )



import json
import logging
from dataclasses import dataclass

from mangum.types import Scope
from mangum.exceptions import WebSocketError, ConfigurationError


import boto3
from botocore.exceptions import ClientError


@dataclass
class WebSocket:

    connection_id: str
    ws_config: dict
    api_gateway_region_name: str
    api_gateway_endpoint_url: str

    def __post_init__(self) -> None:
        self.logger: logging.Logger = logging.getLogger("mangum.websocket")
        try:
            backend = self.ws_config["backend"]
            params = self.ws_config["params"]
        except KeyError as exc:
            raise ConfigurationError(f"WebSocket config {exc} missing.")
        if backend == "sqlite3":
            self.logger.info(
                "The `SQLiteBackend` should be only be used for local "
                "debugging. It will not work in a deployed environment."
            )
            from mangum.backends.sqlite3 import SQLiteBackend

            self._backend = SQLiteBackend(params)  # type: ignore
        elif backend == "dynamodb":
            from mangum.backends.dynamodb import DynamoDBBackend

            self._backend = DynamoDBBackend(params)  # type: ignore
        elif backend == "s3":
            from mangum.backends.s3 import S3Backend

            self._backend = S3Backend(params)  # type: ignore

        elif backend == "postgresql":
            from mangum.backends.postgresql import PostgreSQLBackend

            self._backend = PostgreSQLBackend(params)  # type: ignore

        elif backend == "redis":
            from mangum.backends.redis import RedisBackend

            self._backend = RedisBackend(params)  # type: ignore

        else:
            raise ConfigurationError(f"{backend} is not a supported backend.")

    def create(self, initial_scope: dict) -> None:
        initial_scope_json = json.dumps(initial_scope)
        self._backend.create(self.connection_id, initial_scope_json)

    def fetch(self) -> None:
        initial_scope = self._backend.fetch(self.connection_id)
        scope = json.loads(initial_scope)
        query_string = scope["query_string"]
        headers = scope["headers"]
        if headers:
            headers = [[k.encode(), v.encode()] for k, v in headers.items() if headers]
        scope.update({"headers": headers, "query_string": query_string.encode()})
        self.scope: Scope = scope

    def delete(self) -> None:
        self._backend.delete(self.connection_id)

    def post_to_connection(self, msg_data: bytes) -> None:  # pragma: no cover
        try:
            apigw_client = boto3.client(
                "apigatewaymanagementapi",
                endpoint_url=self.api_gateway_endpoint_url,
                region_name=self.api_gateway_region_name,
            )
            apigw_client.post_to_connection(
                ConnectionId=self.connection_id, Data=msg_data
            )
        except ClientError as exc:
            status_code = exc.response.get("ResponseMetadata", {}).get("HTTPStatusCode")
            if status_code == 410:
                self.delete()
            else:
                raise WebSocketError(exc)



import enum
import asyncio
import logging
from dataclasses import dataclass, field

from mangum.websocket import WebSocket
from mangum.types import ASGIApp, Message


class WebSocketCycleState(enum.Enum):
    REQUEST = enum.auto()
    RESPONSE = enum.auto()


@dataclass
class WebSocketCycle:
    websocket: WebSocket
    log_level: str
    state: WebSocketCycleState = WebSocketCycleState.REQUEST
    response: dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.logger: logging.Logger = logging.getLogger("mangum.asgi.websocket")
        self.loop = asyncio.get_event_loop()
        self.app_queue: asyncio.Queue = asyncio.Queue()
        self.response["statusCode"] = 200
        self.logger.debug("WebSocket cycle initialized!")

    def __call__(self, app: ASGIApp) -> dict:
        asgi_instance = self.run(app)
        asgi_task = self.loop.create_task(asgi_instance)
        self.loop.run_until_complete(asgi_task)

        return self.response

    async def run(self, app: ASGIApp) -> None:
        try:
            await app(self.websocket.scope, self.receive, self.send)
        except BaseException as exc:
            self.logger.error("Exception in ASGI application", exc_info=exc)
            self.response["statusCode"] = 500

    async def receive(self) -> Message:  # pragma: no cover
        message = await self.app_queue.get()

        return message

    async def send(self, message: Message) -> None:
        if self.state is WebSocketCycleState.REQUEST:
            if message["type"] in ("websocket.accept", "websocket.close"):
                self.state = WebSocketCycleState.RESPONSE
            else:
                raise RuntimeError(
                    f"Expected 'websocket.accept' or 'websocket.close', received: {message['type']}"
                )
        else:
            msg_data = message.get("text", "").encode()
            if message["type"] == "websocket.send":
                self.websocket.post_to_connection(msg_data)
            await self.app_queue.put({"type": "websocket.disconnect", "code": "1000"})

    def put_message(self, message: Message) -> None:
        self.app_queue.put_nowait(message)
